/*
 *  HeightField.cpp
 *  Assignment4
 *
 *  Created by Zach Gildersleeve on 2/9/07.
 *  Copyright 2007 __MyCompanyName__. All rights reserved.
 *
 */

#include "HeightField.h"
#include <fstream>
#include <iostream>
#include <float.h>
using namespace std;

HeightField::HeightField(Material *mat, char *name, const Point &newP1, const Point &newP2) : Primitive(mat){

	pt[0] = newP1;
	pt[1] = newP2;
	
   ifstream in(name);
   if(!in){
	   cerr << "Error opening " << name << "\n";
	   exit(1);
   }
   
   double minz, maxz;
   in >> nx >> ny >> minz >> maxz;
   if(!in){
	   cerr << "Error reading header from " << name << "\n";
	   exit(1);
   }
   
   in.get();
   
   data = new float*[nx+1];
   p = new float[(nx+1)*(ny+1)];
   
   for(int i=0;i<=nx;i++)
     data[i] = p+i*(ny+1);
	 
   in.read((char *)p, sizeof(float)*(nx+1)*(ny+1));
   
   if(!in){
     cerr << "Error reading data from " << name << "\n";
     exit(1);
   }
   
   double dz = maxz - minz;
   if (dz < 1.e-3)
	   dz = 1.e-3;
   
   // scale bounds by small amount, as per email
   pt[0] = Point(pt[0].x, pt[0].y, minz - dz * 1.e-6);
   pt[1] = Point(pt[1].x, pt[1].y, maxz + dz * 1.e-6);  
   
   // Step 1
   // D = Pmax = Pmin
   // cellsize = D / (nx, ny, 1)
   diagonal = pt[1] - pt[0];
   cellSize = diagonal / Vector(nx, ny, 1);
   inverseCellSize = cellSize.inverse();
}

bool HeightField::intersect(HitRecord &hit, const RenderContext &context, const Ray &ray) const{
	
	double epsilon = 1e-6;
	
	// STEP 2 INTERSECT WITH BOUNDING BOX, RETURN tNear
	double tmin, tmax, tymin, tymax, tzmin, tzmax, tnear, tfar;
	Vector inverse = ray.inverseDirection;
	
	tmin = (pt[ray.sign.x].x - ray.origin.x) * inverse.x;
	tmax = (pt[1 - ray.sign.x].x - ray.origin.x) * inverse.x;
	tymin = (pt[ray.sign.y].y - ray.origin.y) * inverse.y;
	tymax = (pt[1 - ray.sign.y].y - ray.origin.y) * inverse.y;
	
	if ((tmin > tymax) || (tymin > tmax))
		return false;
	if (tymin > tmin)
	{
		tmin = tymin;
	}
	if (tymax < tmax)
	{
		tmax = tymax;
	}
	
	tzmin = (pt[ray.sign.z].z - ray.origin.z) * inverse.z;
	tzmax = (pt[1 - ray.sign.z].z - ray.origin.z) * inverse.z;
	
	if ((tmin > tzmax) || (tzmin > tmax))
		return false;
	if (tzmin > tmin)
	{
		tmin = tzmin;
	}
	if (tzmax < tmax)
	{
		tmax = tzmax;
	}
		
	if (tmin < epsilon && tmax < epsilon)
		return false;
		
	if (tmin < epsilon)
		tnear = 0.0;
	else
		tnear = tmin;
		
	tfar = tmax;
	
	// step 3 COMPUTER LATTICE COORDINATES OF NEAR POINT
	// WORLD SPACE P = O +tNear * V
	// LATTICE SPACE L = (int)((P - Pmin) / cellsize)
	// watch out for roundoff error
	
	Point P = ray.origin + ray.direction * tnear;
	Vector L = P - pt[0];
		
	int Lx = (int)(L.x * inverseCellSize.x);
	if (Lx < 0)
		Lx = 0;
	else if (Lx >= nx)
		Lx = nx - 1;
	
	int Ly = (int)(L.y * inverseCellSize.y);
	if (Ly < 0)
		Ly = 0;
	else if (Ly >= ny)
		Ly = ny - 1;	
	
	int Lz = 0;
	
	// Step 4 determine how ray marching changes index
	// diy = D.y * V.y > 0 ? 1: -1
	// stopy = D.y * V.y > 0 ? yres : -1
	// same for x
	
	double _x = diagonal.x * ray.direction.x;
	double _y = diagonal.y * ray.direction.y;

	int dix = (_x > 0) ? 1 : -1;
	int stopx = (_x > 0) ? nx : -1;
	
	int diy = (_y > 0) ? 1 : -1;
	int stopy = (_y > 0) ? ny : -1;	
	
	// step 5 determine how t value changes with ray marching
	//dtdx = abs(cellsize.x / V.x)
	//dtdy = abs(cellsize.y / V.y)
	
	double dtdx = fabs(cellSize.x * inverse.x);
	double dtdy = fabs(cellSize.y * inverse.y);
	
	// step 6 determine the far edge of the cell
	// if dix == 1, far.x=(L.x+1)*cellsize.x+Pmin
	// if dix == -1, far.x=L.x*cellsize.x+Pmin
	// same for y
	
	double farX, farY;
	if (dix == 1)
		farX = (Lx + 1) * cellSize.x + pt[0].x;
	else
		farX = Lx * cellSize.x + pt[0].x;
		
	if (diy == 1)
		farY = (Ly + 1) * cellSize.y + pt[0].y;
	else
		farY = Ly * cellSize.y + pt[0].y;
		
	// step 7 determine t values of far slabs
	// tnext_x = (far.x-O.x)/V.x
	// tnext_y = (far.y-O.y)/V.y
	
	double tnext_x = (farX - ray.origin.x) * inverse.x;
	double tnext_y = (farY - ray.origin.y) * inverse.y;
	
	// step 8 beginning of loop
	// compute range of Z values
	// zenter = O.z+tnear*V.z
	// texit = Min(next_x, next_y)
	// zexit = O.z+texit*V.z
	
	double zEnter = ray.origin.z + ray.direction.z * tnear;
	
	while (tnear < tfar)
	{
	
		double tExit = min(tnext_x, tnext_y);
		double zExit = ray.origin.z + tExit * ray.direction.z;
		
		// step 9 determine overlap of z range
		// datamin = Min(data[L.x][L.y], data[L.x+1][L.y], data[L.x][L.y+1], data[L.x+1][L.y+1]) 
		// zmin = Min(zenter, zexit) 
		// Similar for max 
		// if zmin > datamax || zmax<datamin 
			// skip to step 11datamax 
			
		double dataMin = min(min(data[Lx][Ly], data[Lx + 1][Ly]), min(data[Lx][Ly + 1], data[Lx + 1][Ly + 1]));
		double dataMax = max(max(data[Lx][Ly], data[Lx + 1][Ly]), max(data[Lx][Ly + 1], data[Lx + 1][Ly + 1]));
		double zMin = min(zEnter, zExit);
//		zMin -= 0.001;
		double zMax = max(zEnter, zExit);
//		zMax += 0.001;
		
		if (!(zMin > dataMax || zMax < dataMin))
		{
			// step 10 intersect ray with cell
			// using bilinear patch
			
			Point E = ray.origin + ray.direction * tnear;
			Point C = pt[0] + Vector(Lx, Ly, 0) * cellSize; // <------------------- ?
			Vector EC = E - C;
			
			double Sx = EC.x * inverseCellSize.x;
			double Sy = EC.y * inverseCellSize.y;
			
			double Dx = ray.direction.x * inverseCellSize.x;
			double Dy = ray.direction.y * inverseCellSize.y;
			
			double Zc = data[Lx][Ly];
			double Zu = data[Lx + 1][Ly] - Zc;
			double Zv = data[Lx][Ly + 1] - Zc;
			double Zuv = data[Lx + 1][Ly + 1] - data[Lx + 1][Ly] - data[Lx][Ly + 1] + Zc;
			
			double a = Dx * Dy *Zuv;
			double b = (Dx * Zu) + (Dy * Zv) + (Sx * Dy *Zuv) + (Sy * Dx *Zuv) - ray.direction.z;
			double c = Zc + (Sx * Zu) + (Sy * Zv) + (Sx * Sy * Zuv) - E.z;
			
			if (fabs(a) < epsilon)  // check for linear equation
			{
				double tCell = -c / b;
				double t = tCell + tnear;
				if (tCell >= 0 && t <= tExit)
				{
					if (hit.hit(t, this, material))
					{
						// set normal in scratchpad
						
						normalData *nd = hit.getScratchpad<normalData>();
						nd->C = C;
						nd->Zu = Zu;
						nd->Zv = Zv;
						nd->Zuv = Zuv;
						
			//			std::cout << "hit in hf\n";
						return true;
					}
				}
			}
			else // compute quadratic equation
			{
				double discriminant = (b * b) - (4 * a * c);
				if (discriminant > 0)
				{
					double root = sqrt(discriminant);
					double denom = 1.0 / (2.0 * a);
					
					double t1 = (-b + root) * denom;
					double t1pos = t1 + tnear;
					double t2 = (-b - root) * denom;
					double t2pos = t2 + tnear;
					
					if (t1 > 0 && t1pos < tExit)
					{
						if (hit.hit(t1pos, this, material))
						{
							if (t2 > 0)
							{
								hit.hit(t2pos, this, material);
							}
							// set normal
							normalData *nd = hit.getScratchpad<normalData>();
							nd->C = C;
							nd->Zu = Zu;
							nd->Zv = Zv;
							nd->Zuv = Zuv;						
			//										std::cout << "hit in hf\n";
							return true;
						}
					}
					else if (t2 > 0 && t2pos < tExit)
					{
						if (hit.hit(t2pos, this, material))
						{
							// set normal
							normalData *nd = hit.getScratchpad<normalData>();
							nd->C = C;
							nd->Zu = Zu;
							nd->Zv = Zv;
							nd->Zuv = Zuv;					
			//										std::cout << "hit in hf\n";
							return true;
						}
					}
				}
			}
		}
		
		// step 11 March to next cell 
		//if tnext_x < tnext_y 
		//	tnear = tnext_x 
		//	tnext_x += dtdx 
		//	L.x += dix 
		//	else 
		//	Similar for y
		zEnter = zExit;
		
		if (tnext_x < tnext_y)
		{
			Lx += dix;
			if (Lx == stopx)
				break;
			tnear = tnext_x;
			tnext_x += dtdx;
		}
		else
		{
			Ly += diy;
			if (Ly == stopy)
				break;
			tnear = tnext_y;
			tnext_y += dtdy;
		}		
	}
	return false;
}

void HeightField::normal(Vector &normal, const Point &hitPoint, const HitRecord &hit, const RenderContext &context, const Ray &ray) const{

	const normalData *nd = hit.getScratchpad<normalData>();
	Point C = nd->C;
	double Zu = nd->Zu;
	double Zv = nd->Zv;
	double Zuv = nd->Zuv;
	
	double u = (hitPoint.x - C.x) * inverseCellSize.x;
	double v = (hitPoint.y - C.y) * inverseCellSize.y;		
	double nx = (Zu + v * Zuv) * inverseCellSize.x;
	double ny = (Zv + u * Zuv) * inverseCellSize.y;
	
	normal = Vector(-nx, -ny, 1.0);
	normal.normalize();


}

void HeightField::getBounds(BoundingBox& bbox) const {
	bbox.build(pt[0]);
	bbox.build(pt[1]);	
}



